# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse
from astropy.io import fits
import numpy as np

from azulero import stats
from azulero.timing import Timer


def add_parser(subparsers):
    parser = subparsers.add_parser(
        "tune",
        help="Autotune image rendering parameters.",
        description=("Analyse image statistics to propose a white point."),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "filename",
        type=str,
        metavar="PATH",
        help="Path to the input data",
    )
    parser.add_argument(
        "--zero",
        type=float,
        default=24.5,
        metavar="ZP",
        help="Zero point",
    )

    parser.set_defaults(func=run)


def run(args):

    timer = Timer()

    print(f"Read input: {args.filename}")
    with fits.open(args.filename) as f:
        data = f[0].data
    print(f"- Shape: {data.shape[0]} x {data.shape[1]}")
    timer.tic_print()

    propose_white_point(data, args.zero)  # FIXME split
    timer.tic_print()


def propose_white_point(data: np.ndarray, zp: float):

    print(f"Compute image statistics:")
    qs = [0, 0.01, 0.1, 1, 50, 99, 99.9, 99.99, 100]
    percentiles = stats.percentiles(data[data > 0], qs)
    percentiles.values = -2.5 * np.log10(percentiles.values) + zp
    for q in percentiles:
        print(f"- {q}: {percentiles[q]}")

    white = percentiles[99.99]

    print(f"Clipping adjustment:")
    clipping = white - percentiles[100]
    print(f"- Base white point: {white}")
    print(f"- Max: {percentiles[100]}")
    print(f"- Clipping: {clipping}")
    if clipping > 1.0:
        adj = -min(clipping * 0.3, 1.5)
        print(f"- Adjustment: {adj}")
        white += adj

    print(f"Saturation adjustment:")
    sat_frac = np.sum(data > 0.9 * percentiles[99.9]) / data.size
    print(f"- Saturated fraction: {sat_frac}")
    if sat_frac > 0.001:
        adj = -min(sat_frac * 300, 1.5)
        print(f"- Adjustment: {adj}")
        white += adj

    print(f"Dynamic range ajustment:")
    dr = percentiles[1] - percentiles[99]
    print(f"- Dynamic range: {dr}")
    if dr > 10:
        adj = (10 - dr) * 0.12
        print(f"- High DR adjustment: {adj}")
        white += adj
    elif dr < 8.75:
        adj = (8.75 - dr) * 0.08
        print(f"- Low DR adjustment: {adj}")
        white += adj

    print(f"Clip at zero point:")
    print(f"- Zero point: {zp}")
    print(f"- White point: {white}")
    return min(white, zp)



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from azulero import color


def test_scaling():

    data = np.ones((2, 4, 3), dtype=int) * 12
    raw = data.view(np.ma.MaskedArray)

    raw = color.channelwise_div(raw, (3, 2))

    assert np.all(raw[0] == 4)
    assert np.all(raw[1] == 6)



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import time
import tracemalloc


class Timer(object):  # FIXME rename as profiler

    def __init__(self):
        tracemalloc.start()
        self.start = time.perf_counter()
        self.prev = self.start
        self.split = 0
        self.total = 0
        self.size = 0
        self.peak = 0

    def tic(self):
        prev = self.prev
        self.prev = time.perf_counter()
        self.split = self.prev - prev
        self.total = self.prev - self.start
        self.size, self.peak = tracemalloc.get_traced_memory()
        self.size /= 1024**3
        self.peak /= 1024**3
        tracemalloc.reset_peak()

    def tic_print(self):
        self.tic()
        print(f"- Elapsed time: {self.split:.1f}s / Total: {self.total:.1f}s")
        print(f"- Memory usage: {self.size:.1f}GB / Peak: {self.peak:.1f}GB")



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import enum
import numpy as np
from skimage.restoration import inpaint as skinpaint
import cv2


class Flag(enum.Enum):

    @classmethod
    def valid(cls, value):
        for flag in cls:
            if value & 2**flag.value:
                # print(f"{value} & {flag.value} ({flag.name})")
                return False
        return True

    @classmethod
    def invalid(cls, value):
        return not cls.valid(value)


class VisFlag(Flag):
    HOT = 0
    COLD = 1
    SATURATED = 2
    BAD = 8


class NirFlag(Flag):
    INVALID = 1
    DISCONNECTED = 2
    ZERO_QE = 3
    SUPER_QE = 6
    HOT = 7
    SNOWBALL = 9
    SATURATED = 10
    NL_SATURATED = 12


def dead_pixels(iyjh):
    return iyjh == 0


def hot_pixels(i, y, j, h):

    abs_threshold = 10.0
    rel_threshold = 10.0
    hot_i = (i > abs_threshold) & (i > rel_threshold * h)
    hot_y = (y > abs_threshold) & (y > rel_threshold * j)
    hot_j = (j > abs_threshold) & (j > rel_threshold * h)
    hot_h = (h > abs_threshold) & (h > rel_threshold * y)
    return hot_i | hot_y | hot_j | hot_h


def inpaint(data: np.ndarray, mask: np.ndarray, axis: int = -1):
    if data.ndim > 2:
        return skinpaint.inpaint_biharmonic(data, mask, channel_axis=axis)
    return cv2.inpaint(data, mask.astype(np.uint8), 3, cv2.INPAINT_NS)
    # return skinpaint.inpaint_biharmonic(data, mask)


def _resaturate(x):
    if x <= 0.8:
        return x
    if x >= 0.9:
        return 1
    return 2 * x - 0.8


def resaturate(data, factor=1.0):
    if len(data) == 0:
        return data
    return np.vectorize(_resaturate)(data / factor) * factor



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse

from azulero import assemble, find, retrieve, crop, tune, process, roam


def run():

    parser = argparse.ArgumentParser(
        prog="azul",
        description="Bring colors to Euclid tiles!",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--workspace", type=str, default=".", metavar="PATH", help="Parent workspace"
    )
    parser.add_argument(
        "--input",
        "-i",
        type=str,
        default="*[-_]{channel}[-_]*.fits",
        metavar="PATTERN",
        help="Input file pattern, where `{channel}` is replaced with the channel name, e.g. `NIR-Y`",
    )

    subparsers = parser.add_subparsers(title="Commands")
    find.add_parser(subparsers)
    retrieve.add_parser(subparsers)
    crop.add_parser(subparsers)
    tune.add_parser(subparsers)
    process.add_parser(subparsers)
    assemble.add_parser(subparsers)
    roam.add_parser(subparsers)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    run()



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

from astropy.io import fits
import numpy as np
import tifffile
from pathlib import Path


def parse_tile(text: str):
    tile_slicing = text.split("[")
    if len(tile_slicing) == 1:
        return tile_slicing[0], None
    return tile_slicing[0], parse_slice(tile_slicing[1][:-1])


def parse_slice(text: str):
    """
    Parse a 2D slice, e.g. ":,3:14".
    """
    if text is None:
        return None
    parse_index = lambda i: int(i) if i else None
    return tuple(
        slice(*[parse_index(i) for i in axis.split(":")]) for axis in text.split(",")
    )


def parse_map(text: str, dtype=float):
    """
    Parse a comma-separated list of 'key:value' pairs.
    """
    if not text:
        return []
    pairs = [p.split(":") for p in text.split(",")]
    return [[dtype(x), dtype(y)] for x, y in pairs]


def read_fits(path: Path, slicing=None):
    """
    Read a region in the primary array of a FITS file.
    """
    data = fits.getdata(path)
    return data if slicing is None else data[slicing]


def make_workdir(workspace, tile):
    workdir = Path(workspace).expanduser() / tile
    if workdir.is_dir():
        print("WARNING: Working directory already exists.")
    else:
        workdir.mkdir(parents=True)
    return workdir


def write_fits(data: np.array, path: Path):
    """
    Write an SIF file.
    """
    fits.PrimaryHDU(data).writeto(path, overwrite=True)


def write_tiff(rgb: np.ndarray, path: Path):
    """
    Write a normalized RGB image.
    """
    data = np.flipud(rgb * 65535).astype(np.uint16)
    tifffile.imwrite(path, data)


def write_mask(iyjh: np.ndarray, path: Path):
    """
    Write a 4-channel binary mask.
    """
    i, y, j, h = iyjh
    rgb = np.zeros((iyjh.shape[1], iyjh.shape[2], 3), dtype=np.uint8)
    rgb[:, :, 0] = i * 155 + h * 100
    rgb[:, :, 1] = i * 155 + j * 100
    rgb[:, :, 2] = i * 155 + y * 100
    tifffile.imwrite(path, np.flipud(rgb))


def read_channel(workdir: Path, pattern: str, slicing=None):
    """
    Read the region of one channel.
    """
    data_files = list(workdir.glob(pattern))

    if len(data_files) == 1:
        return read_fits(data_files[0], slicing)

    print(f"WARNING: {len(data_files)} files found with pattern: {pattern}")
    return _average([read_fits(f, slicing) for f in data_files])


def _average(slices: list):
    """
    Average arrays, discarding zeros.
    """
    stack = np.stack(slices)
    # FIXME replace 0 with np.nan
    return np.nan_to_num(
        np.nanmean(stack, axis=0, where=(stack != 0)), copy=False
    )  # FIXME nanmedian


def read_iyjh(workdir: Path, slicing=None, template="{}"):
    """
    Read the region of a VIS- and NIR-covered tile.
    """
    return np.stack(
        (
            read_channel(workdir, template.format(channel="VIS"), slicing),
            read_channel(workdir, template.format(channel="NIR-Y"), slicing),
            read_channel(workdir, template.format(channel="NIR-J"), slicing),
            read_channel(workdir, template.format(channel="NIR-H"), slicing),
        )
    )



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
import numpy as np
import cv2
from scipy import interpolate
from skimage.filters import unsharp_mask as sksharpen

from azulero import io  # FIXME rm


@dataclass
class Transform(object):
    iyjh_zero_points: np.ndarray
    iyjh_scaling: np.ndarray
    iyjh_fwhm: np.ndarray
    sharpen_strength: float
    nir_to_l: float
    i_to_b: float
    y_to_g: float
    j_to_r: float
    hue: float
    saturation: float
    stretch: float
    bw: np.ndarray


def sharpen(data, radii, strength):  # FIXME to dedicated module
    if strength == 0:
        return data
    for i in range(len(data)):
        data[i] = sksharpen(data[i], radii[i], strength, True)
    return data


def abmag_to_value(mag, zp):
    return 10 ** ((zp - mag) / 2.5)


def stretch_iyjh(data: np.ndarray, transform: Transform):
    whites = abmag_to_value(transform.bw[1], transform.iyjh_zero_points)
    scaling = (transform.iyjh_scaling / whites)[:, np.newaxis, np.newaxis]
    a = abmag_to_value(transform.bw[1], transform.stretch)
    b = -abmag_to_value(transform.bw[0], transform.bw[1])
    data *= scaling
    return asinh(data, a, b)


def iyjh_to_lrgb(data: np.ndarray, transform: Transform):
    i, y, j, h = data
    lrgb = np.zeros((data.shape[1], data.shape[2], 4), dtype=np.float32)
    lrgb[:, :, 0] = lerp(transform.nir_to_l, np.median(data[1:], axis=0), data[0])
    lrgb[:, :, 1] = lerp(transform.j_to_r, j, h)
    lrgb[:, :, 2] = lerp(transform.y_to_g, y, j)
    lrgb[:, :, 3] = lerp(transform.i_to_b, i, y)
    return lrgb


def lrgb_to_rgb(data: np.ndarray, transform: Transform):
    hls = cv2.cvtColor(data[:, :, 1:], cv2.COLOR_RGB2HLS)
    hls[:, :, 0] = (hls[:, :, 0] + transform.hue) % 360
    hls[:, :, 2] = np.clip(hls[:, :, 2] * transform.saturation, 0, 1)
    hls[:, :, 1] = data[:, :, 0]
    return cv2.cvtColor(hls, cv2.COLOR_HLS2RGB)


def lerp(x, a, b):
    if x == 0:
        return b
    if x == 1:
        return a
    return x * a + (1 - x) * b


def channelwise_mul(data, factors):
    for i in range(len(factors)):
        data[i] = data[i] * factors[i]
    return data


def channelwise_div(data, factors):
    for i in range(len(factors)):
        data[i] = data[i] / factors[i]
    return data


def asinh(data: np.ndarray, a: float, black: float):
    b = np.arcsinh(black * a)
    data *= a
    np.arcsinh(data, out=data)
    data -= b
    data /= np.arcsinh(a) - b
    np.clip(data, 0, 1, out=data)
    return data


def adjust_curve(data: np.ndarray, knots: list):
    x, y = list(map(list, zip(*knots)))
    k = min(len(knots) - 1, 3)
    spline = interpolate.make_interp_spline(x, y, k)
    return np.clip(spline(data), 0.0, 1.0, dtype=np.float32)  # FIXME inplace



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from azulero import stats


def test_percentiles():

    a = np.arange(-42, 10001)
    np.random.shuffle(a)
    qs = [0, 0.01, 0.05, 0.5, 1, 50, 99, 99.9, 99.99, 100]
    p = stats.percentiles(a[a >= 0], qs)
    expected = [0, 1, 5, 50, 100, 5000, 9900, 9990, 9999, 10000]
    assert np.allclose(p.values, expected)
    for i in range(len(qs)):
        assert p[qs[i]] == expected[i]



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from azulero import mask


def test_flagging():

    assert mask.VisFlag.valid(0)
    assert mask.VisFlag.invalid(1)
    assert mask.VisFlag.invalid(2)
    assert mask.VisFlag.invalid(3)
    assert mask.VisFlag.invalid(4)
    assert mask.VisFlag.valid(8)

    assert mask.NirFlag.valid(0)
    assert mask.NirFlag.valid(1)
    assert mask.NirFlag.invalid(2)
    assert mask.NirFlag.invalid(3)
    assert mask.NirFlag.invalid(4)
    assert mask.NirFlag.invalid(8)
    assert mask.NirFlag.valid(16)
    assert mask.NirFlag.invalid(2**6)
    assert mask.NirFlag.invalid(2**7)
    assert mask.NirFlag.invalid(2**9)
    assert mask.NirFlag.invalid(2**10)
    assert mask.NirFlag.valid(2**11)
    assert mask.NirFlag.invalid(2**12)


def test_inpainting():

    data = np.ones((9, 16, 3))
    flags = np.zeros((9, 16))
    data[1, 1, 1] = 0
    flags[1, 1] = 1

    res = mask.inpaint(data, flags)

    assert np.all(res == 1)



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse
import numpy as np
from pathlib import Path
import cv2
import yaml

from azulero import sequence
from azulero.timing import Timer


def add_parser(subparsers):
    parser = subparsers.add_parser(
        "roam",
        help="Create a video which pans and zooms in an image.",
        description=(
            "Supply an image, specify viewport position and parameters at given times."
            "They will be interpolated to render a smooth roaming video."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "input",
        type=str,
        metavar="FILENAME",
        help="Input image file.",
    )
    parser.add_argument(
        "sequence",
        type=str,
        metavar="FILENAME",
        help="YAML configuration file which specifies the sequence of key frames.",
    )
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        default="output.mp4",
        metavar="FILENAME",
        help="Output video file.",
    )
    parser.add_argument(
        "--format",
        type=int,
        nargs=2,
        default=[1920, 1080],
        metavar=["WIDTH", "HEIGHT"],
        help="Video format",
    )
    parser.add_argument(
        "--fps", type=float, default=24, metavar="FPS", help="Frames per second."
    )

    parser.set_defaults(func=run)


def run(args):

    print()

    input = Path(args.workspace).expanduser() / args.input
    config = Path(args.sequence)
    output = Path(args.output)

    timer = Timer()

    print(f"Read input image: {input}")
    image = cv2.imread(input, cv2.IMREAD_COLOR)
    image_shape = image.shape[:2]
    print(f"- Shape: {image_shape[0]} x {image_shape[1]}")
    timer.tic_print()

    print(f"Read sequence of key frames: {config}")
    with open(config) as f:
        params = sequence.load_frames_params(
            yaml.safe_load(f), image_shape, args.fps, args.format
        )
    print(f"- Key frames: {len(params)}")
    centers = sequence.sin_sequence(params.centers)
    zooms_inv = sequence.sin_sequence(params.zooms_inv)
    angles_deg = sequence.sin_sequence(params.angles_deg)
    print(f"- Total frames: {len(centers)}")
    timer.tic_print()

    print(f"Generate frames")
    writer = cv2.VideoWriter(
        output, cv2.VideoWriter_fourcc(*"mp4v"), args.fps, args.format
    )

    print(f"- Frame\tx\ty\tz (%)\ta (°)")
    for f, c, z, a in zip(range(len(centers)), centers, zooms_inv, angles_deg):
        print(f"- {f}\t{c[0]:0.1f}\t{c[1]:0.1f}\t{100.0/z:0.1f}\t{a:0.1f}")
        p = sequence.KeyFrame(0, c, 1.0 / z, a)
        frame = crop(image, p, args.format)
        writer.write(frame)

    writer.release()
    print(f"- Output written: {output}")
    timer.tic_print()


def crop(image: np.ndarray, params, format):
    center = params.center
    viewport_format = np.array(format, dtype=float) / params.z
    a = params.a_deg % 360
    viewport = cv2.RotatedRect(params.center, viewport_format, -a)
    x0, y0, w, h = viewport.boundingRect()
    vertical = w < h
    if vertical:
        # OpenCV unhappy!
        x0, y0, w, h = y0, x0, h, w
        image = np.swapaxes(image, 0, 1)
        a = 90 - a
        center = np.flip(center)
    x1 = x0 + w
    y1 = y0 + h
    if x0 < 0:
        print(f"- WARNING: min(x) < 0 ({x0})")
        x0 = 0
    if y0 < 0:
        print(f"- WARNING: max(y) min < 0 ({y0})")
        y0 = 0
    if x1 > image.shape[1]:
        print(f"- WARNING: max(x) > {image.shape[1]-1} ({x1-1})")
        x1 = image.shape[0]
    if y1 > image.shape[0]:
        print(f"- WARNING: max(y) > {image.shape[0]-1} ({y1-1})")
        y1 = image.shape[1]
    offset = np.array([x0, y0])
    patch = image[y0:y1, x0:x1]
    rotation = cv2.getRotationMatrix2D(center - offset, a, params.z)
    rotation_format = (w, h)
    rotated_image = cv2.warpAffine(
        patch, rotation, rotation_format, flags=cv2.INTER_LINEAR
    )
    res = cv2.getRectSubPix(rotated_image, format, center - offset)
    if vertical:
        return np.flipud(res)
    return res



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from azulero import sequence


def test_coord_parsing():

    assert sequence.parse_coord("10.5px", 100) == 10.5
    assert sequence.parse_coord("-10.5px", 100) == 89.5
    assert sequence.parse_coord("10.5%", 200) == 21.0
    assert sequence.parse_coord("-10.5%", 200) == 179.0


def test_zoom_parsing():

    assert sequence.parse_zoom("20%", None, None) == 0.2
    assert sequence.parse_zoom("0.2w", [160, 180], [16, 9]) == 0.5
    assert sequence.parse_zoom("0.2h", [160, 180], [16, 9]) == 0.25


def test_angle_parsing():

    assert sequence.parse_a_deg("20°") == 20
    assert sequence.parse_a_deg("-0.5pi") == -90


def test_trajectory_sampling():

    start_frame = 10
    stop_frame = 17
    start_centers = [[0, 1], [1, 2], [2, 3]]
    stop_centers = [3, 4]
    centers = sequence.sin_spline(
        sequence.KeyValue(start_frame, start_centers),
        sequence.KeyValue(stop_frame, stop_centers),
    )
    assert np.allclose([c[1] for c in centers], [c[0] + 1 for c in centers])
    for c0, c1 in zip(centers[:-1], centers[1:]):
        assert c1[0] > c0[0]
        assert c1[1] > c0[1]



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse
import math
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path

from azulero import io
from azulero.timing import Timer


def add_parser(subparsers):
    parser = subparsers.add_parser(
        "crop",
        help="Display the VIS channel for region selection.",
        description=(
            "Display the VIS channel between values 0 and 1 in a new window, "
            "and enable cropping a region by zooming in. "
            "When the window is closed, "
            "the program prints out the corresponding image processing command."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "tile",
        type=str,
        metavar="INDEX",
        help="Tile index.",
    )
    parser.add_argument(
        "--white", "-w", type=float, default=1.0, metavar="VALUE", help="White point"
    )
    parser.add_argument(
        "--round",
        type=int,
        default=500,
        metavar="INC",
        help=(
            "Image region rounding increment. "
            "The returned region is the smallest region which contains the selected region, "
            "and whose bounds are multiples of the increment."
        ),
    )

    parser.set_defaults(func=run)


def run(args):

    print()

    workdir = Path(args.workspace).expanduser() / args.tile

    timer = Timer()

    print(f"Read VIS channel: {workdir}")
    data = io.read_channel(workdir, args.input.format("VIS"))
    timer.tic_print()

    print(f"Prepare data.")
    h, w = data.shape
    data = np.clip(data[::10, ::10], 0, args.white)
    im = plt.imshow(np.flipud(np.asinh(data / 0.7)), extent=[0, w, 0, h])
    timer.tic_print()

    plt.title("Zoom to select a region, then close the window.")
    plt.show()

    rounding = args.round
    x0, x1 = im.axes.get_xlim()
    y0, y1 = im.axes.get_ylim()
    x0 = math.floor(x0 / rounding) * rounding
    x1 = min(math.ceil(x1 / rounding) * rounding, w)
    y0 = math.floor(y0 / rounding) * rounding
    y1 = min(math.ceil(y1 / rounding) * rounding, h)

    print(f"\nYou may now run:")
    print(
        f"\nazul --workspace {args.workspace} process {args.tile}[{y0}:{y1},{x0}:{x1}]\n"
    )



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse
from astropy.coordinates import SkyCoord
from dataclasses import dataclass
import json
import pathlib
from shapely import geometry

from azulero.timing import Timer


def add_parser(subparsers):

    parser = subparsers.add_parser(
        "find",
        help="Find the tiles which contain objects.",
        description="Find object coordinates and intersecting tiles.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "objects",
        type=str,
        nargs="*",
        metavar="NAMES",
        help="Object names.",
    )
    parser.add_argument(
        "--radec",
        type=str,
        nargs=2,
        action="append",
        metavar=["RA", "DEC"],
        help="Coordinates (this option can be specified several times).",
    )
    parser.add_argument(
        "--tiling",
        type=str,
        default="DpdMerFinalCatalog.geojson",
        metavar="FILENAME",
        help="Geojson file which lists existing tiles and their metadata",
    )

    parser.set_defaults(func=run)


@dataclass
class Tile(object):
    index: int
    mode: str
    dsr: str
    distance: float

    def __str__(self) -> str:
        return f"{self.mode}: {self.index} ({self.dsr}); distance: {self.distance:.2f}°"


class Tiling(object):

    def __init__(self, filename):
        print(f"Load tiling: {filename}")
        with open(filename) as f:
            self.tiles = json.load(f)["features"]
        print(f"- {len(self.tiles)} tiles loaded.")

    def __call__(self, coord: SkyCoord):
        matches = {}
        point = geometry.Point(coord.ra.degree, coord.dec.degree)
        for tile in self.tiles:
            polygon = geometry.shape(tile["geometry"])
            if polygon.contains(point):
                # FIXME use astropy-region for spherical geometry
                index = tile["properties"]["TileIndex"]
                mode = tile["properties"]["ProcessingMode"]
                dsr = tile["properties"]["DatasetRelease"]
                center = polygon.centroid
                distance = center.distance(point)
                if distance < 1:
                    matches[index] = Tile(index, mode, dsr, distance)
                    # FIXME avoid overwriting dsr
        return sorted(matches.values(), key=lambda t: t.distance)


def run(args):

    print()

    timer = Timer()
    tiling = Tiling(pathlib.Path(args.workspace) / args.tiling)
    timer.tic_print()

    print()

    objects = args.objects + [SkyCoord(*rd, unit="deg") for rd in args.radec]
    for coord in objects:
        if isinstance(coord, str):
            print(coord)
            coord = SkyCoord.from_name(coord)
            print(f"- Coordinates: {coord.ra:.2f}, {coord.dec:.2f}")
        else:
            print(f"{coord.ra.degree} {coord.dec.degree}")
        tiles = tiling(coord)
        for t in tiles:
            print(f"- {t}")
        timer.tic_print()
        if len(tiles) > 0:
            print(f"\nYou may now run:")
            print(
                f"\nazul --workspace {args.workspace} retrieve {' '.join(str(t.index) for t in tiles)}\n"
            )
        else:
            print("\nWARNING: No tile found.\n")



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
import scipy.interpolate as interp
import numpy as np

from azulero import color  # TODO lerp to interp.py


@dataclass
class KeyFrame:
    frame: int
    center: np.ndarray
    z: float
    a_deg: float

    def __repr__(self) -> str:
        return f"{self.frame}: ({self.center[0]:0.1f}, {self.center[1]:0.1f}), {int(self.z * 100+0.5)}%, {self.a_deg}°"


@dataclass
class KeyValue:
    frame: int
    value: object


@dataclass
class KeyFrames:
    centers: list
    zooms_inv: list
    angles_deg: list

    def __len__(self):
        return len(self.centers)

    def append(self, frame, center, zoom, angle):
        if center[0] is None or center[1] is None:  # FIXME can there be a single None?
            self.centers.append(KeyValue(frame, self.centers[-1].value))
        else:
            self.centers.append(KeyValue(frame, center))
        if zoom is None:
            self.zooms_inv.append(KeyValue(frame, self.zooms_inv[-1].value))
        elif not np.isnan(zoom):
            self.zooms_inv.append(KeyValue(frame, 1.0 / zoom))
        if angle is None:
            self.angles_deg.append(KeyValue(frame, self.angles_deg[-1].value))
        elif not np.isnan(angle):
            self.angles_deg.append(KeyValue(frame, angle))
        return self


def load_frames_params(
    sequence: list, image_shape: list, fps: float, video_format: list
):
    res = KeyFrames([], [], [])
    frame = 0
    for step in sequence:
        if not "t" in step:
            x = parse_coord(step["x"], image_shape[1])
            y = parse_coord(step["y"], image_shape[0])
            add_knot(res, x, y)
        else:
            frame = parse_frame(step["t"], fps, frame)
            x = None if "x" not in step else parse_coord(step["x"], image_shape[1])
            y = None if "y" not in step else parse_coord(step["y"], image_shape[0])
            z = (
                None
                if "z" not in step
                else parse_zoom(step["z"], image_shape, video_format)
            )
            a = None if "a" not in step else parse_a_deg(step["a"])
            res.append(frame, np.array([x, y]), z, a)
    return res


def add_knot(sequence, x, y):
    knots = sequence.centers[-1].value
    if isinstance(knots, list):
        sequence.centers[-1].value.append(np.array([x, y]))
    else:
        sequence.centers[-1].value = [knots, np.array([x, y])]


def sin_sequence(keys_values: list):
    """
    Interpolate parameters over a sequence of frames with sine sampling.
    """
    res = []
    for start, stop in zip(keys_values[:-1], keys_values[1:]):
        if isinstance(start.value, list):
            res += [*sin_spline(start, stop)]
        else:
            res += sin_step(start, stop)
    # FIXME prepend first value if first frame > 0
    return res


def sin_step(start, stop):
    """
    Linearly interpolate parameters between two frames with sine sampling.
    """
    stop_value = stop.value if not isinstance(stop.value, list) else stop.value[0]
    return [
        color.lerp(1 - u, start.value, stop_value)
        for u in sin_sampling(start.frame, stop.frame)
    ]


def sin_spline(start, stop):
    """
    Spline-interpolate trajectory between knots with sine sampling.
    """
    knots = np.stack([*start.value, stop.value])
    b = interp.make_interp_spline(
        np.linspace(0, 1, len(knots)), knots, k=min(3, len(knots) - 1)
    )
    u = sin_sampling(start.frame, stop.frame)
    return b(u)


def sin_sampling(start, stop):
    """
    Sine sampling between two bounds.
    """
    return np.sin(np.linspace(0, 1, stop - start) * np.pi - np.pi / 2) / 2 + 0.5


def parse_frame(text: str, fps: float, ref_frame: int):
    """
    Parse frame index or time.
    If last char is "f", return the value.
    If it is "s", multiply by `fps`.
    If the first char is "+", add `ref_frame`.
    """
    if text[-1] == "f":
        value = int(text[:-1])
    elif text[-1] == "s":
        value = int(float(text[:-1]) * fps)
    else:
        raise ValueError(f"Unrecognized time: {text}")
    return value + ref_frame if text[0] == "+" else value


def parse_coord(text: str, image_extent: int):
    """
    Parse a coordinate.
    If last char is "%", coordinate is relative to the image extent.
    If value is negative, index backward.
    """
    if text.endswith("px"):
        px = float(text[:-2])
    elif text[-1] == "%":
        px = float(text[:-1]) / 100 * image_extent
    else:
        ValueError(f"Unrecognized coordinate: {text}")
    if px < 0:
        px += image_extent
    return px


def parse_zoom(text: str, image_shape: list, video_format: list):
    """
    Parse the zoom.
    If last char is "w" (resp. "h"), zoom is relative to the image width (resp. height).
    If last char is "%", zoom is a relative to the pixel size.
    """
    if text == "...":
        return np.nan
    if text[-1] == "w":
        z = video_format[0] / image_shape[1] / float(text[:-1])
    elif text[-1] == "h":
        z = video_format[1] / image_shape[0] / float(text[:-1])
    elif text[-1] == "%":
        z = float(text[:-1]) / 100
    else:
        raise ValueError(f"Unrecognized zoom: {text}")
    return z


def parse_a_deg(text: str):
    """
    Parse the angle in degrees.
    If last char is "°", forward the value.
    If text ends with "pi", multiply by 180.
    """
    if text == "...":
        return np.nan
    if text[-1] == "°":
        return float(text[:-1])
    elif text.endswith("pi"):
        return float(text[:-2]) * 180
    raise ValueError(f"Unrecognized angle: {text}")



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from azulero import io


def test_slicing():

    text = ":,3:14"
    slicing = io.parse_slice(text)
    assert slicing == (slice(None, None), slice(3, 14))
    a = np.zeros((9, 16))
    b = a[slicing]
    assert b.shape == (9, 11)



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
import numpy as np


@dataclass
class KeysValues:
    """
    Ordered, possibly-float indexed, dict.
    """

    keys: list
    values: list

    def __iter__(self):
        return iter(self.keys)

    def __getitem__(self, q):
        i = self.keys.index(q)
        return self.values[i]

    def __repr__(self) -> str:
        return (
            "{" + ", ".join(f"{k}: {v}" for k, v in zip(self.keys, self.values)) + "}"
        )


def percentiles(data: np.ndarray, qs: list):
    """
    Compute a list of percentiles without interpolation.
    """
    sorted = data.flatten()
    last = len(sorted) - 1
    sorted.sort()  # TODO partial sort upto `round(max(qs) / 100 * last)`
    return KeysValues(qs, [sorted[round(q / 100 * last)] for q in qs])



# SPDX-FileCopyrightText: Copyright (C) 2025, Antoine Basset
# SPDX-PackageSourceInfo: https://github.com/kabasset/azulero
# SPDX-License-Identifier: Apache-2.0

import argparse
import numpy as np
from pathlib import Path

from azulero import color, io, mask
from azulero.timing import Timer


def add_parser(subparsers):
    parser = subparsers.add_parser(
        "process",
        help="Process MER channels to render a color image.",
        description=(
            "Process MER channels: "
            "1. Inpaint dead pixels; "
            "2. Sharpen IYJH channels; "
            "3. Stretch dynamic range with asinh function; "
            "4. Blend IYJH channels into RGB and lightness (L) channels; "
            "5. Shift hue and boost color saturation; "
            "6. Adjust curves."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "tile",
        type=str,
        metavar="SPEC",
        help="Tile index and optional slicing à-la NumPy, e.g. 102160611[1500:7500,11500:17500]",
    )
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        default="{tile}_{step}.tiff",
        metavar="TEMPLATE",
        help=(
            "Output filename or template, relative to the tile folder. "
            "Placeholder {tile} is replaced by the tile index, "
            "and {step} is replaced by the processing step. "
            "If {step} is not present in the template, "
            "intermediate steps are not saved."
        ),
    )
    parser.add_argument(
        "--zero",
        nargs=4,
        type=float,
        default=[24.5, 29.8, 30.1, 30.0],
        metavar=("ZP_I", "ZP_Y", "ZP_J", "ZP_H"),
        help="Zero points for each band",
    )
    parser.add_argument(
        "--scaling",
        nargs=4,
        type=float,
        default=[
            2.2,
            1.3,
            1.2,
            1.0,
        ],
        metavar=("GAIN_I", "GAIN_Y", "GAIN_J", "GAIN_H"),
        help="Scaling factors applied immediately to the IYJH bands for white balance",
    )
    parser.add_argument(
        "--fwhm",
        nargs=4,
        type=float,
        default=[1.6, 3.5, 3.4, 3.5],
        metavar=("FWHM_I", "FWHM_Y", "FWHM_J", "FWHM_H"),
        help="FWHM for each band",
    )
    parser.add_argument(
        "--sharpen",
        type=float,
        default=0.5,
        metavar="STRENGTH",
        help="Strength of the sharpening",
    )
    parser.add_argument(
        "--nirl",
        type=float,
        default=0.1,
        metavar="RATE",
        help="NIR contribution to L, between 0 and 1.",
    )
    parser.add_argument(
        "--ib",
        type=float,
        default=1.0,
        metavar="RATE",
        help="I contribution to B, between 0 and 1.",
    )
    parser.add_argument(
        "--yg",
        type=float,
        default=0.5,
        metavar="RATE",
        help="Y contribution to G, between 0 and 1.",
    )
    parser.add_argument(
        "--jr",
        type=float,
        default=0.25,
        metavar="RATE",
        help="J contribution to R, between 0 and 1.",
    )
    parser.add_argument(
        "--white",
        "-w",
        type=float,
        default=22.0,
        metavar="VALUE",
        help="White point in AB magnitude.",
    )
    parser.add_argument(
        "--stretch",
        "-a",
        type=float,
        default=28.0,
        metavar="FACTOR",
        help="Stretching factor in AB magnitude.",
    )
    parser.add_argument(
        "--offset",
        "-b",
        type=float,
        default=29.0,
        metavar="VALUE",
        help="Opposite of black point in AB magnitude.",
    )
    parser.add_argument(
        "--hue", type=float, default=-20, metavar="ANGLE", help="Hue shift"
    )
    parser.add_argument(
        "--saturation",
        type=float,
        default=1.2,
        metavar="GAIN",
        help="Saturation factor",
    )
    parser.add_argument(
        "--curves",
        type=str,
        nargs=3,
        default=["", "", "0.5: 0.55"],
        metavar=("KNOTS_R", "KNOTS_G", "KNOTS_B"),
        help="Curve spline knots for each channel",
    )

    parser.set_defaults(func=run)


def run(args):

    print()

    transform = color.Transform(
        iyjh_zero_points=np.array(args.zero),
        iyjh_scaling=np.array(args.scaling),
        iyjh_fwhm=np.array(args.fwhm),
        sharpen_strength=args.sharpen,
        nir_to_l=args.nirl,
        i_to_b=args.ib,
        y_to_g=args.yg,
        j_to_r=args.jr,
        hue=args.hue,
        saturation=args.saturation,
        stretch=args.stretch,
        bw=np.array([args.offset, args.white]),
    )

    tile, slicing = io.parse_tile(args.tile)
    workdir = Path(args.workspace).expanduser() / tile
    name = args.output.replace("{tile}", tile)

    timer = Timer()

    print(f"Read IYJH image from: {workdir}")
    iyjh = io.read_iyjh(workdir, slicing, args.input)
    print(f"- Shape: {iyjh.shape[1]} x {iyjh.shape[2]}")
    timer.tic_print()

    print(f"Detect bad pixels")
    dead = mask.dead_pixels(iyjh)
    print(f"- Dead pixels: {', '.join(str(np.sum(channel)) for channel in dead)}")
    if "{step}" in name:
        path = workdir / name.replace("{step}", "mask")
        print(f"- Write: {path.name}")
        io.write_mask(dead, path)
    timer.tic_print()

    print(f"Inpaint dead pixels")
    iyjh[0] = mask.inpaint(iyjh[0], dead[0])
    # iyjh[0][dead[0]] = mask.resaturate(iyjh[0][dead[0]], np.max(iyjh[0]))
    nir_dead = dead[1] | dead[2] | dead[3]
    iyjh[1:] = mask.inpaint(iyjh[1:], nir_dead, 0)
    timer.tic_print()

    print(f"Sharpen channels")
    iyjh = color.sharpen(iyjh, transform.iyjh_fwhm / 2.355, transform.sharpen_strength)
    timer.tic_print()

    print(f"Stretch dynamic range")
    iyjh = color.stretch_iyjh(iyjh, transform)
    # print(f"- Medians: {', '.join(str(np.median(c)) for c in iyjh)}") # TODO use approximant
    timer.tic_print()
    # TODO save vstacked iyjh (crop if too high)

    print(f"Blend IYJH to RGB")
    lrgb = color.iyjh_to_lrgb(iyjh, transform)
    del iyjh
    rgb = color.lrgb_to_rgb(lrgb, transform)
    del lrgb
    if "{step}" in name:
        path = workdir / name.replace("{step}", "blended")
        print(f"- Write: {path.name}")
        io.write_tiff(rgb, path)
    timer.tic_print()

    # print(f"Inpaint hot pixels")
    # rgb[dead[0]] = mask.resaturate(rgb[dead[0]])
    # rgb = mask.inpaint(rgb, hot)
    # timer.tic_print()

    # if "{step}" in name:
    #     path = workdir / name.replace("{step}", "inpainted")
    #     print(f"- Write: {path.name}")
    #     io.write_tiff(rgb, path)
    #     timer.tic_print()

    print(f"Adjust curves")
    for i in range(len(args.curves)):
        knots = io.parse_map(args.curves[i])
        knots.insert(0, [0, 0])
        knots.append([1, 1])
        rgb[:, :, i] = color.adjust_curve(rgb[:, :, i], knots)
    timer.tic_print()

    path = workdir / name.replace("{step}", "adjusted")
    print(f"- Write: {path.name}")
    io.write_tiff(rgb, path)
    timer.tic_print()
